package prupp; 

// $Id: HTTPHeader.java,v 1.4 1999/01/17 22:01:42 d94-rol Exp $

import java.io.*;
import java.util.*;

/** Reads in a http-header (either a response or a request)
 *  and sets up for some nice handling of the header.
 */
public class HTTPHeader 
{
    private String line;                      // the request- or statusline
    private String method = "";
    private String requestURI = "";
    private String httpVersion = "";
    private Vector headers = new Vector ();     // the headers in order
    private Hashtable htab = new Hashtable ();  // to quikly find/set headers.
    private static final String CRLF = "\r\n";
    
    class Header 
    {
	String type;
	String lowercase;
	String value;
	Header (String type, String value) {
	    this.type = type;
	    this.value = value;
	    this.lowercase = type.toLowerCase ();
	}
	public boolean equals (Object o) {
	    if (o instanceof Header) {
		return (((Header)o).lowercase.equals (lowercase));
	    }
	    return false;
	}
    }
    
    /** Create a new HTTPHeader from scratch
     */
    public HTTPHeader () {    
    }
    
    /** Read in a HTTPHeader from an inputstream.
     * @param in the stream from which this header is read.
     */
    public HTTPHeader (DataInputStream in) throws IOException {
	String line = in.readLine ();
	if (line == null)
	    throw (new IOException ("Couldnt read requestline"));
	setRequestLine (line);

	String msg;
	while (true) {
	    msg = in.readLine ();
	    if (msg == null) 
		throw (new IOException ("Couldnt read headers, connection must be closed"));
	    if (msg.equals (""))
		break;
	    int i = msg.indexOf (':');
	    if (i < 0)
		throw (new IOException ("Malformed header:" + msg));
	    Header head = new Header (msg.substring (0,i), msg.substring (i+2));
	    headers.addElement (head);
	    htab.put (head.lowercase, head);
	}
    }
    /** get the text value of this header 
     * @return a String describing this HTTP-Header.
     */
    public String toString () {
	StringBuffer ret = new StringBuffer (getRequestLine ());
	ret.append (CRLF);
	for (int i = 0; i < headers.size (); i++) {
	    Header h = (Header)headers.elementAt (i);
	    ret.append (h.type + ": " + h.value + CRLF);
	}
	ret.append (CRLF);
	return ret.toString ();
    }
    
    /** get the value of header type 
     * @param type the Header were intrested in.
     * @return the value of type or null if no value is set.
     */
    public String getHeader (String type) {
	Header h = (Header)htab.get (type.toLowerCase ());
	if (h != null)
	    return (h.value);
	return null;
    }
    /** Set or replaces a value for given type.
     * @param type the type or category that we want to set.
     * @param value the value we want to set
     */
    public void setHeader (String type, String value) {
	Header h = new Header (type, value);
	    
	Header hold = (Header)htab.get (h.lowercase);
	if (hold != null) {
	    int i = headers.indexOf (h);
	    headers.setElementAt (h, i);
	    htab.put (h.lowercase, h);
	} else {
	    headers.addElement (h);
	    htab.put (h.lowercase, h);
	}
    }

    /** removes a headerline from this header
     * @param type the type we want to remove
     */
    public void removeHeader (String type) {
	Header h = new Header (type, null);
	headers.removeElement (h);
	htab.remove (h.lowercase);
    }

    /** Get the statusline of this header (only valid for responses). 
     * @return the status of the request.
     */
    public String getStatusline () {
	return getRequestLine ();
    }
    
    /** Set the statusline of this header.
     * @param line a Status-Line )RFC 2068: 6.1)
     */
    public void setStatusLine (String line) {
	setRequestLine (line);
    }

    /** Get the requestline of this header (only valid for requests).
     * @return the request. 
     */
    public String getRequestLine () {
	return method + ' ' + requestURI + ' ' + httpVersion;
    }

    /** Set the requestline of this header
     * @param line a Request-Line (RFC 2068: 5.1)
     */
    public void setRequestLine (String line) {
	this.line = line;
	int s1 = line.indexOf (' ');
	if (s1 < 0) {
	    method = line;
	    return;
	}
	int s2 = line.indexOf (' ', s1+1);
	method = line.substring(0,s1);
	if (s2 > 0) {
	    requestURI = line.substring (s1+1,s2);
	    httpVersion = line.substring (s2+1);
	}
    }

    /** Is this request for the head only?
     * @return true if this request is for HEAD, false otherwise
     */
    public boolean isHeadOnlyRequest () {
	return method.equals ("HEAD");     // method is casesensitive.
    }
    
    /** Get the request method of this header (only valid for requests).
     * @return the request method.
     */
    public String getMethod (){
	return method;
    }
    
    /** Sets the request method of this header 
     * @param method the new requestmethod
     */
    public void setMehtod (String method) {
	this.method = method;
    }

    /** Get the requestURI of this request (only valid for requests).
     * @return the requestURI.
     */
    public String getRequestURI () {
	return requestURI;
    }
    
    /** Sets the request URI of this header 
     * @param method the new URI
     */
    public void setRequestURI (String requestURI) {
	this.requestURI = requestURI;
    }
    
    /** Get the HTTP Version of this request (only valid for requests).
     * @return the http version.
     */
    public String getHTTPVersion () {
	return httpVersion;
    }

    /** Set the HTTP Version to use for request.
     * @param version the version to use.
     */
    public void setHTTPVersion (String version) {
	httpVersion = version;
    }
    
    /** Get the HTTP version of the response (only valid for responses).
     * @return the HTTP version.
     */
    public String getResponseHTTPVersion () {
	return method;
    }
    
    /** Set the HTTP version for this response.
     * @param httpVersion the version to use.
     */
    public void setResponseHTTPVersion (String httpVersion) {
	method = httpVersion;
    }

    /** Get the Status code of the response (only valid for responses).
     * @return the status code.
     */
    public String getStatusCode () {
	return requestURI;
    }
    
    /** Set the Status code for this response.
     * @param status the new status code.
     */
    public void setStatusCode (String status) {
	requestURI = status;
    }

    /** Get the Reason phrase of the response (only valid for responses).
     * @return the reason phrase.
     */
    public String getReasonPhrase () {
	return httpVersion;
    }
    
    /** Set the reason phrase for this reqponse.
     * @param reason the new reasonphrase
     */
    public void setReasonPhrase (String reason) {
	httpVersion = reason;
    }
}


