package prupp; 

// $Id: HTTPHeaderHandler.java,v 1.16 1999/01/17 22:01:42 d94-rol Exp $

import java.net.*;
import java.io.*;
import java.util.*;
import java.util.zip.*;


/** Do some magic.<pre>
 * Handles the (client)requst to the Proxy, figures out what to do. 
 *
 * A typical request:
 *
 * GET /index.html HTTP/1.0
 * Proxy-Connection: Keep-Alive
 * User-Agent: Mozilla/4.04j2 [en] (X11; I; SunOS 5.5.1 sun4u)
 * Pragma: no-cache
 * Host: www.ludd.luth.se
 * Accept: image/gif, image/x-xbitmap, image/jpeg, image/pjpeg, image/png
 * Accept-Language: en
 * Accept-Charset: iso-8859-1,*,utf-8
 * </pre><br><pre>
 * SPEC from HTTP-1.1 (RFC 2068)
 * Request        = Request-Line              ; Section 5.1
 * *( general-header         ; Section 4.5
 * | request-header         ; Section 5.3
 * | entity-header )        ; Section 7.1
 * CRLF
 * [ message-body ]          ; Section 7.2
 * Request-Line   = Method SP Request-URI SP HTTP-Version CRLF
 * </pre>
 */
public class HTTPHeaderHandler extends Thread
{
    private URL url;
    private DataOutputStream out;
    private DataInputStream in;
    private Socket socket;
    private String content = null;
    private Responder responder;

    /* Metapages */
    /** URL of nonfiltered page */
    private static final String NOPROXY="http://noproxy.";
    /** is this a noproxy-request. */
    private boolean noproxy = false;
    /** URL of META-pages */
    private static final String META="http://proxy/";
    /** is this a meta-page. */
    private boolean meta = false;
  
    /** are we allowed to use the cache? */
    protected boolean useCached = true;

    /** are we allowed to cache the page? */
    protected boolean cacheAllowed = true;

    /** the current request */
    protected String request;

    /** current status of connection */
    private String statusline;
    /** request were handling */
    private String connection;
    /** time we started processing */
    private Date started;

    /** reads the HTTP Header from the socket and determine what to do. 
     * @param socket the client were talking to.
     */
    public HTTPHeaderHandler (Socket socket) {
	this.socket = socket;
	Proxy.getLogger().inc ("HTTPHeaderHandler(s) created");
	connection = "";
	statusline = "Created";
	started = new Date ();
	this.start ();
    }

    /** Reads the request and finds out which type of handler to use 
     *  tries to use the cache if possible and if not use a webreader to 
     *  handle the fetching and sending of the file.
     */
    public void run () {
	statusline = "Started";
	try {
	    in = new DataInputStream (socket.getInputStream ());
	    out = new DataOutputStream (socket.getOutputStream ());
	} catch (IOException e) {
	    try {
		socket.close ();
		socket = null;
	    } catch (IOException e2) {
		Proxy.logError ("Something is really bad right now");
		System.exit (-4711);
	    }
	    Proxy.logError ("Couldnt get streams");
	    cleanup ();
	    return;
	}
    
	statusline = "Reading request";
	String requestLine;
	HTTPHeader header = null;
	try {
	    header = new HTTPHeader (in);
	} catch (IOException e) {
	    Proxy.logError ("Connection " + socket.getInetAddress ().getHostAddress() + " prematurly terminated:" + e);
	    cleanup ();	    
	    return;
	}
	
	requestLine = header.getRequestLine ();
	responder = new Responder (requestLine, in, out); // create a fake one.
	
	if (!setURL (header)) {
	    cleanup ();	    
	    return;
	}
	
	if (meta) {
	    if (Proxy.allowed (socket, Access.ADMIN))	
		responder = new MetaPageHandler (socket, requestLine, in, out, url, header);
	    else {
		responder.forbidden (url);
		Proxy.logError ("" + socket + " trying to enter a restricted area");
	    }
	    cleanup ();
	    return;
	}
    
	if (!Proxy.allowed (socket, Access.HTTP)) {
	    responder.forbidden (url);
	    Proxy.logError ("" + socket + " trying to enter a restricted area");
	    cleanup ();
	    return;      
	}   

	header.removeHeader ("Proxy-authorization");
	if (noproxy) {
	    useCached = false;
	    String host = header.getHeader ("Host");
	    if (host != null) {
		host = host.substring ("noproxy.".length());
		header.setHeader ("Host", host);
	    }
	}

	if (header.getHeader ("Content-length") != null) {
	    useCached = false;   // dont cache post-results.
	    if (!readContent (header)) {
		responder.setStatusCode (400);
		cleanup ();
		return;
	    }
	}
	
	// standard web-authentication.
	if (header.getHeader ("authorization") != null || 
	    header.getHeader ("proxy-authorization") != null ||
	    (header.getHeader ("cookie") != null &&   // cookie-passwords suck.
	     ((header.getHeader ("cookie").toLowerCase ().indexOf ("password") >= 0) ||
	      (header.getHeader ("cookie").toLowerCase ().indexOf ("id") >= 0)))) {
	    useCached = false;     // dont use cached files, 
	    cacheAllowed = false;  // and dont cache it.
	}	
	
	String cached = header.getHeader ("Pragma");
	if (cached != null && cached.trim().equals ("no-cache")) {
	    Proxy.getLogger().inc ("Pragma no-cache");
	    useCached = false;
	}
	cached = header.getHeader ("Cache-Control");
	if (cached != null && cached.trim().equals ("no-cache")) {
	    Proxy.getLogger().inc ("Pragma no-cache");
	    useCached = false;
	}
	    
	statusline = "Sending File";

	if (header.isHeadOnlyRequest ()) {
	    useCached = false;
	    cacheAllowed = false;
	}

	try {
	    if (useCached) {
		NCacheEntry entry = Proxy.getCache ().getEntry (url);
		if (entry != null) {
		    Proxy.getLogger ().inc ("Cache hits");
		    responder = new CacheSender (requestLine, in, out, entry);
		} else {
		    Proxy.getLogger ().inc ("Cache misses");
		    responder = new WebReader (requestLine, 
					       in, 
					       out, 
					       url, 
					       header.toString () + (content != null ? content : "" ), 
					       noproxy,
					       cacheAllowed);
		}
	    } else {
		responder = new WebReader (requestLine, 
					   in, 
					   out, 
					   url, 
					   header.toString () + (content != null ? content : "" ), 
					   noproxy,
					   cacheAllowed);
	    }
	} catch (IOException e) {
	    Proxy.logError ("Responding failed: " + e);
	    responder.setStatusCode (500);
	} catch (Exception e) {
	    StringWriter sw  = new StringWriter ();
	    PrintWriter pw = new PrintWriter (sw);
	    e.printStackTrace (pw);	   
	    pw.close ();
	    Proxy.logError ("We have a new bug to handle:\n" + sw);	   
	    responder.internalError (sw.toString ());
	}
	cleanup ();
    }

    /** remove the remaining resources.
     */
    private void cleanup () {
	if (responder != null)
	    Proxy.logRequest (socket, responder);
	Proxy.removeConnection (this);
	try {
	    out.close ();
	    in.close ();
	    socket.close ();
	} catch (IOException e) {
	    Proxy.logError ("Having a hard time closing the connection: " + e);
	}
    }    

    /** If this request has a body (due to some post or so)
     *  read it in and append it to the request.
     */
    protected boolean readContent (HTTPHeader header) {
	String sl = (String)header.getHeader("Content-length");
	int contentLength = Integer.parseInt (sl.trim ());

	byte v[] = new byte[contentLength];
	try {
	    in.readFully (v, 0, contentLength);   // well we should read this many!
	    content = new String (v);
	} catch (EOFException e) {
	    Proxy.logError ("Couldnt read all content");
	    return false;
	} catch (IOException e) {
	    Proxy.logError ("Unexpected termination:" + e);
	    return false;
	}
	return true;
    }
  
    /** Try to make an URL of the request. Can also be some special page
     *  that this proxy provides (like proxy.status).
     *  This little thingy sets us up for correctness when the user is
     *  requesting a Meta-page (a page the proxy provides, like proxy.status).
     *  @param header the HTTP header for this request.
     */
    protected boolean setURL(HTTPHeader header) {
	String uri = header.getRequestURI ();

	if (uri.toLowerCase ().startsWith (NOPROXY)) {
	    uri = "http://" + uri.substring (NOPROXY.length ());
	    noproxy = true;
	}
	if (uri.toLowerCase ().startsWith (META)) {
	    meta = true;
	}
    
	try {
	    connection = uri;
	    url = new URL (uri);
	    try {
		InetAddress urlhost = Proxy.getInetAddress (url);
		if (urlhost != null && urlhost.equals (Proxy.getHost ()))
		    meta = true;
	    } catch (UnknownHostException e) {
	    }
	} catch (MalformedURLException e) {
	    Proxy.logError ("Bad URL in request: " + header.getRequestURI ());
	    return false;
	}

	if (Proxy.isProxyConnected ())
	    header.setRequestURI (uri);
	
	else {
	    if (uri.indexOf ("/",uri.indexOf ("://")+3) == -1)
		uri = uri + "/";
	    header.setRequestURI (uri.substring (uri.indexOf ("/",uri.indexOf ("://")+3)));  // Request-URI	    
	}
	return true;
    }

    /** Get the current status of this request so we can build status-pages etc.
     * @return the current status of this connection.
     */
    public String getStatus () 
	{
	    return statusline;
	}
    /** Get the connection of this request for status-pages etc.
     * @return the current connection.
     */
    public String getConnection () 
	{
	    return connection;
	}

    /** Get the time this connection started for status-pages etc.
     * @return the Date this connections started
     */
    public Date getStarted () 
	{
	    return started;
	}

    /** Get the InetAddress of user
     */
    public InetAddress getInetAddress () {
	return socket.getInetAddress ();
    }
}

